"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var RuntimeEnvironment_1 = require("./RuntimeEnvironment");
var Generic_1 = require("./Generic");
var DirectMap_1 = require("./DirectMap");
var DirectTree;
(function (DirectTree) {
    DirectTree.DEFAULT_SEPERATOR = '/';
    /**
     * Traverse the tree to apply action.
     */
    function TraverseLeaf(directTree, action, currentPath, seperator) {
        if (currentPath === void 0) { currentPath = ''; }
        if (seperator === void 0) { seperator = DirectTree.DEFAULT_SEPERATOR; }
        if (Generic_1.Generic.IsObject(directTree)) {
            var prefix_1 = currentPath.length > 0 ? "" + currentPath + seperator : '';
            Object.keys(directTree).forEach(function (key) {
                TraverseLeaf(directTree[key], action, "" + prefix_1 + key, seperator);
            });
        }
        else {
            action(directTree, currentPath);
        }
    }
    /**
     * The reverse-action of Flat, parse a flat tree and build a normal tree.
     * @param seperator seperator used in path
     */
    function Deflat(flatTree, seperator) {
        if (seperator === void 0) { seperator = DirectTree.DEFAULT_SEPERATOR; }
        var result = {};
        Object.keys(flatTree).forEach(function (key) {
            var path = seperator !== DirectTree.DEFAULT_SEPERATOR ? key.split(seperator).join(DirectTree.DEFAULT_SEPERATOR) : key;
            Set(result, path, flatTree[key]);
        });
        return result;
    }
    DirectTree.Deflat = Deflat;
    /**
     * Delete() deletes the subtree with provided path.
     */
    function Delete(directTree, path) {
        var branch = GetBranch(directTree, path);
        if (branch) {
            RuntimeEnvironment_1.RuntimeEnvironment.Delete(branch.parent, branch.key);
        }
    }
    DirectTree.Delete = Delete;
    /**
     * Create a flat direct map of leaves with each leaf's path (or path's variant) as it's new key.
     * @param seperator seperator used in path
     */
    function Flat(directTree, seperator) {
        if (seperator === void 0) { seperator = DirectTree.DEFAULT_SEPERATOR; }
        var flat = {};
        TraverseLeaf(directTree, function (value, path) {
            flat[path] = value;
        }, '', seperator);
        return flat;
    }
    DirectTree.Flat = Flat;
    /**
     * GetBranch() gets the parent of provided path and it's branch key.
     */
    function GetBranch(directTree, path) {
        if (directTree[path] !== undefined) {
            return { parent: directTree, key: path };
        }
        else {
            var keys = path.split(DirectTree.DEFAULT_SEPERATOR);
            // tslint:disable:no-empty
            for (; keys.length > 1 && directTree; directTree = directTree[keys.splice(0, 1)[0]]) { }
            return keys.length === 1 && directTree ? { parent: directTree, key: keys[0] } : null;
        }
    }
    DirectTree.GetBranch = GetBranch;
    /**
     * Get() returns the subtree or leaf with provided path.
     */
    function Get(directTree, path) {
        if (path && path.length > 0) {
            var branch = GetBranch(directTree, path);
            return branch ? branch.parent[branch.key] : undefined;
        }
        else
            return directTree;
    }
    DirectTree.Get = Get;
    /** IsFlat() determines whether a tree is flat */
    function IsFlat(directTree) {
        return Object.values(directTree).every(function (value) { return !Generic_1.Generic.IsObject(value); });
    }
    DirectTree.IsFlat = IsFlat;
    /**
     * MergeAt() merges another tree into tree's branch.
     */
    function MergeAt(directTree, path, anotherTree) {
        var branch = GetBranch(directTree, path);
        if (branch != null) {
            var target = branch.parent[branch.key];
            if (Generic_1.Generic.IsObject(target)) {
                MergeInternal(target, anotherTree);
            }
            else {
                MergeInternal(branch.parent, DirectMap_1.DirectMap.Pack(branch.key, anotherTree));
            }
        }
    }
    DirectTree.MergeAt = MergeAt;
    /**
     * MergeExternal() merges tree and another tree into a new tree.
     */
    function MergeExternal(directTree, anotherTree) {
        var newTree = Generic_1.Generic.Clone(directTree);
        MergeInternal(newTree, anotherTree);
        return newTree;
    }
    DirectTree.MergeExternal = MergeExternal;
    /**
     * MergeInternal() merges anotherTree into tree.
     */
    function MergeInternal(directTree, anotherTree) {
        Object.keys(anotherTree).forEach(function (path) {
            var branch = Get(directTree, path);
            var anotherBranch = anotherTree[path];
            if (Generic_1.Generic.IsPlainObject(branch) && Generic_1.Generic.IsPlainObject(anotherBranch)) {
                MergeInternal(branch, anotherBranch);
            }
            else {
                Set(directTree, path, anotherBranch);
            }
        });
    }
    DirectTree.MergeInternal = MergeInternal;
    function OperateLeaf(outTree, directTree, leafOperator) {
        var operandTrees = [];
        for (var _i = 3; _i < arguments.length; _i++) {
            operandTrees[_i - 3] = arguments[_i];
        }
        Object.keys(directTree).forEach(function (key) {
            if (Generic_1.Generic.IsObject(directTree[key])) {
                OperateLeaf.apply(void 0, [outTree[key],
                    directTree[key],
                    leafOperator].concat(operandTrees.map(function (t) { return t[key]; })));
            }
            else {
                RuntimeEnvironment_1.RuntimeEnvironment.Set(outTree, key, leafOperator.apply(void 0, [directTree[key]].concat(operandTrees.map(function (t) { return t[key]; }))));
            }
        });
    }
    /**
     * OperateLeafInternal() applys provided operator on each leaf value and save the result to a new tree.
     */
    function OperateLeafExternal(directTree, leafOperator) {
        var operandTrees = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            operandTrees[_i - 2] = arguments[_i];
        }
        var newTree = Generic_1.Generic.Null(directTree);
        OperateLeaf.apply(void 0, [newTree, directTree, leafOperator].concat(operandTrees));
        return newTree;
    }
    DirectTree.OperateLeafExternal = OperateLeafExternal;
    /**
     * OperateLeafInternal() applys provided operator on each leaf value directly in source tree.
     */
    function OperateLeafInternal(directTree, leafOperator) {
        var operandTrees = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            operandTrees[_i - 2] = arguments[_i];
        }
        OperateLeaf.apply(void 0, [directTree, directTree, leafOperator].concat(operandTrees));
    }
    DirectTree.OperateLeafInternal = OperateLeafInternal;
    /**
     * Set() sets the subtree with provided path.
     */
    function Set(directTree, path, value) {
        var keys = path.split(DirectTree.DEFAULT_SEPERATOR);
        for (var i = 0; i < keys.length - 1; ++i) {
            var key = keys[i];
            if (directTree[key] == null) {
                directTree[key] = {};
            }
            directTree = directTree[key];
        }
        RuntimeEnvironment_1.RuntimeEnvironment.Set(directTree, keys[keys.length - 1], value);
    }
    DirectTree.Set = Set;
    /**
     * Swap branch of two raw tree.
     */
    function Swap(t1, t2, path) {
        var s1 = Get(t1, path);
        var s2 = Get(t2, path);
        if (s1 != null && s2 != null) {
            Set(t1, path, s2);
            Set(t2, path, s1);
            return true;
        }
        else
            return false;
    }
    DirectTree.Swap = Swap;
})(DirectTree = exports.DirectTree || (exports.DirectTree = {}));
//# sourceMappingURL=DirectTree.js.map